;;;; This file is part of XMLTools
;;;; Copyright (C) 2004, 2007, 2008, 2015 Sergey Poznyakoff
;;;;
;;;; XMLTools is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 3 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; XMLTools is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with XMLTools.  If not, see <http://www.gnu.org/licenses/>.
;;;;

(define-module (xmltools xmltrans)
  :export (xmltrans:parse-error
	   xmltrans:parse-warning
	   xmltrans:parent?
	   xmltrans:get-parent
	   xmltrans:expect-next-tag)
  :export-syntax (xmltrans:make-tag
		  xmltrans:tag-ref
		  xmltrans:tag-set!
		  xmltrans:start-tag
		  xmltrans:end-tag
		  xmltrans:replace-tag))

(use-modules (gamma expat)
	     (ice-9 rdelim))

(if (= (string->number (major-version)) 1)
    (use-syntax (ice-9 syncase)))

(define *debug-level* #f)

(define-public xmltrans:version "2.0-pre")
(define-public xmltrans:version-string
  (string-append "xmlt version "
		 xmltrans:version
		 " ($Id: xmltrans.scm 153 2015-01-11 15:23:51Z gray $)"))

(define-public xmltrans:license
  "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
This is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law.\n\
\n\
Written by Sergey Poznyakoff.\n")

(define-public xmltrans:copyright
  "Copyright (C) 2004, 2007, 2008, 2010, 2015 Sergey Poznyakoff")

(define-public xmltrans:generator
  (string-append
   "XMLTools version "
   xmltrans:version
   " "
   xmltrans:copyright))



;;;; Debug calls
(define indent 0)
(define (indent-increase)
  (set! indent (1+ indent)))
(define (indent-decrease)
  (set! indent (- indent 1)))

(define (trace indent . rest)
  (if *debug-level*
      (begin
	(do ((i 0 (1+ i)))
	    ((= i indent) #f)
	  (display " "))
	(for-each
	 (lambda (text)
	   (display text))
	 rest)
	(newline))))

(define (debug level . rest)
  (if (and *debug-level* (>= *debug-level* level))
      (begin
	(for-each
	 display
	 rest)
	(newline))))

(define-public (xmltrans:set-debug-level level)
  (if (or (not level) (number? level))
      (set! *debug-level* level)
      (throw 'misc-error 'xmltrans:set-debug-level
	     "Level has invalid type."
	     (list) #f)))


;;;; Handler table
;;;; Each element is a cons:
;;;;     (TAG-NAME . #(START-HANDLER END-HANDLER))
;;;;
;;;; (define (START-HANDLER name attrs
;;;;  ... )
;;;;
;;;; Returns text
;;;;
;;;; (define (END-HANDLER name attrs text)
;;;;  ... )
;;;;
;;;; Returns: text string
;;;;        | #f meaning cancel the tag without outputting anything
;;;;
(define handler-table '())

(define default-start-handler
  (lambda (name attrs)
    ""))

(define default-end-handler
  (lambda (name attrs text)
    text))

(define-public (xmltrans:get-default-start-handler)
  default-start-handler)

(define-public (xmltrans:set-default-start-handler proc)
  (set! default-start-handler proc))

(define-public (xmltrans:get-default-end-handler)
  default-end-handler)

(define-public (xmltrans:set-default-end-handler proc)
  (set! default-end-handler proc))

(define-syntax handler-start
  (syntax-rules ()
    ((handler-start x) (or (vector-ref x 0) default-start-handler))))

(define-syntax handler-start-set!
  (syntax-rules ()
    ((handler-start-set! x v) (vector-set! x 0 v))))

(define-syntax handler-end
  (syntax-rules ()
    ((handler-end x) (or (vector-ref x 1) default-end-handler))))

(define-syntax handler-end-set!
  (syntax-rules ()
    ((handler-end-set! x v) (vector-set! x 1 v))))

(define (handle-start name attrs)
  (let ((elem (assoc name handler-table)))
    (debug 2 "Calling START handler for " name ". Found " elem)
    (if elem
	((handler-start (cdr elem)) name attrs)
	(default-start-handler name attrs))))

(define (handle-end name attrs text)
  (let ((elem (assoc name handler-table)))
    (debug 2 "Calling END handler for " name ". Found " elem)
    (if elem
	((handler-end (cdr elem)) name attrs text)
	(default-end-handler name attrs text))))

(define (handler-find-or-insert tag)
  (or (assoc tag handler-table)
      (begin
	(set! handler-table (cons
			     (cons tag (vector #f #f))
			     handler-table))
	(car handler-table))))

(define-public (xmltrans:set-start-handler tag handler)
  (let ((elem (handler-find-or-insert tag)))
    (handler-start-set! (cdr elem) handler)))

(define-public (xmltrans:set-end-handler tag handler)
  (let ((elem (handler-find-or-insert tag)))
    (handler-end-set! (cdr elem) handler)))




;;;; Syntactic sugar

(define-syntax xmltrans:start-tag
  (syntax-rules ()
    ((xmltrans:set-start-handler tag-name arg ...)
     (letrec ((proc (lambda arg
		      ...)))
       (xmltrans:set-start-handler tag-name proc)))))

(define-syntax xmltrans:end-tag
  (syntax-rules ()
    ((xmltrans:set-end-handler tag-name arg ...)
     (letrec ((proc (lambda arg
		      ...)))
       (xmltrans:set-end-handler tag-name proc)))))

(define-syntax xmltrans:replace-tag
  (syntax-rules ()
    ((xmltrans:replace-tag old-tag new-tag)
     (xmltrans:end-tag
      (symbol->string 'old-tag)
      (tag attrs text)
      (string-append "<" (symbol->string 'new-tag) ">" text "</" (symbol->string 'new-tag) ">")))))



;;;; Text stack
;;;; Each element is a cons:
;;;;     (TAG-NAME . #(ATTRIBUTES TEXT-LIST))
;;;; where TAG-NAME is name of the tag, ATTRIBUTES is a list of attributes
;;;; specified with the opening tag, TEXT-LIST is a list of text strings
;;;; collected so far for this tag.

(define-syntax xmltrans:make-tag
  (syntax-rules ()
    ((xmltrans:make-tag name)
     (cons name (vector '() '())))
    ((xmltrans:make-tag name attr)
     (cons name (vector attr '())))
    ((xmltrans:make-tag name attr text)
     (cons name (vector attr text)))))

(define-syntax xmltrans:tag-ref
  (syntax-rules ()
    ((xmltrans:tag-ref #:name x) (car x))
    ((xmltrans:tag-ref #:attrs x) (vector-ref (cdr x) 0))
    ((xmltrans:tag-ref #:text x) (vector-ref (cdr x) 1))))

(define-syntax xmltrans:tag-set!
  (syntax-rules ()
    ((xmltrans:tag-set! #:name x v) (set-car! x v))
    ((xmltrans:tag-set! #:attrs x v) (vector-set! (cdr x) 0 v))
    ((xmltrans:tag-set! #:text x v) (vector-set! (cdr x) 1 v))))

(define tag-list '())

(define (push-tag name attrs)
  (set! tag-list (cons (xmltrans:make-tag name attrs) tag-list))
  (push-text (handle-start name attrs)))

(define (preprocess-text tag)
  (reverse (xmltrans:tag-ref #:text tag)))

(define (pop-tag)
  (debug 10 "pop-tag " (xmltrans:tag-ref #:name (car tag-list)))
  (let* ((tag (car tag-list))
	 (retval (handle-end (xmltrans:tag-ref #:name tag)
			     (xmltrans:tag-ref #:attrs tag)
			     (apply string-append
				    (preprocess-text tag)))))
    (set! tag-list (cdr tag-list))
    (cond
     ((not retval)
      "")
     ((or (list? retval) (string? retval))
      (debug 10 "popped text " retval) 
      retval)
     (else
      (throw 'misc-error 'pop-tag
	     "End handler for element ~A returned unexpected value: ~S"
	     (list (xmltrans:tag-ref #:name tag) retval) #f)))))

(define (push-text text)
  (debug 10 "push-text " text)
  (cond
   ((null? tag-list)) ;; FIXME Ignore text
   ((or (not text) (null? text)))     ;; Ignore text
   ((pair? text)
    (push-text (car text))
    (push-text (cdr text)))
   ((string? text)
    (xmltrans:tag-set! #:text (car tag-list)
	      (cons text (xmltrans:tag-ref #:text (car tag-list)))))
   (else
    (throw 'misc-error 'push-text
	   "Start handler for element ~A returned unexpected value: ~S"
	   (list (xmltrans:tag-ref #:name (car tag-list)) text) #f))))

;;;; For each tag in the tag stack execute the given PROC
;;;; PROC takes two arguments: tag name and a list of attributes
;;;; associated with it.
(define-public (xmltrans:for-each-tag proc)
  (for-each
   (lambda (t)
     (proc (xmltrans:tag-ref #:name t)
	   (xmltrans:tag-ref #:attrs t)))
   tag-list))

;;;; Returns #t if TAG exists on stack
(define-public (xmltrans:tag-exists? tag)
  (assoc tag tag-list))

;;;; Returns attributes associated with the given TAG. TAG must currently
;;;; be open, otherwise exception is thrown.
;;;;   TAG type         returns value of
;;;; =================================================
;;;;    string          the topmost tag with this name
;;;;      #f            the topmost tag on stack
;;;;      #t            the next after the topmost tag on stack
;;;;    number          Nth tag on stack
;;;;

(define-public (xmltrans:extract-item tlist elem)
  (cond
   ((not elem)
    (car tlist))
   ((string? elem)
    (assoc elem tlist))
   ((boolean? elem)
    (if (not (null? (cdr tlist)))
	(cadr tlist)
	#f))
   ((number? elem)
    (if (< elem (length tlist))
	(list-ref tlist elem)
	#f))
   (else
    #f)))

(define-public (xmltrans:tag-attributes tag)
  (if (list? tag)
      tag
      (let ((item (xmltrans:extract-item tag-list tag)))
	(if (not item)
	    (throw 'misc-error 'xmltrans:tag-attributes
		   "No such tag on stack ~S" (list tag) #f))
	(xmltrans:tag-ref #:attrs item))))

;;;; Return attribute NAME associated with the TAG. TAG must currently
;;;; be open, otherwise exception is thrown.
;;;; If TAG is #f, the topmost tag on stack is used
;;;; If TAG has no attribute NAME #f is returned
(define-public (xmltrans:attr tag name)
  (let ((attr (assoc name (xmltrans:tag-attributes tag))))
    (and attr (cdr attr))))

(define-public (xmltrans:set-attr tag name value)
  (let ((item (xmltrans:extract-item tag-list tag)))
    (if (not item)
	(throw 'misc-error 'xmltrans:set-attr
	       "No such tag on stack ~S" (list tag) #f))
    (let ((attr (assoc name (xmltrans:tag-ref #:attrs item))))
      (if attr
	  (set-cdr! attr value)
	  (xmltrans:tag-set! #:attrs
			     item
			     (cons
			      (cons name value)
			      (xmltrans:tag-ref #:attrs item)))))))

(define (xmltrans:parent? name . rest)
  (let ((this-tag (and (not (null? rest)) (car rest))))
    (cond
     ((string? this-tag)
      (call-with-current-continuation
       (lambda (exit)
	 (do ((cur tag-list (cdr cur)))
	     ((null? cur) #f)
	   (if (string=? (xmltrans:tag-ref #:name (car cur)) this-tag)
	       (exit (string=? (xmltrans:tag-ref #:name (cadr cur)) name)))))))
     ((pair? this-tag)
      (call-with-current-continuation
       (lambda (exit)
	 (do ((cur tag-list (cdr cur)))
	     ((null? cur) #f)
	   (if (eq? (car cur) this-tag)
	     (exit (string=? (xmltrans:tag-ref #:name (cadr cur)) name)))))))
     (else
      (string=? (xmltrans:tag-ref #:name (cadr tag-list)) name)))))

(define (xmltrans:get-parent . rest)
  (let ((this-tag (and (not (null? rest)) (car rest))))
    (cond
     ((string? this-tag)
      (call-with-current-continuation
       (lambda (exit)
	 (do ((cur tag-list (cdr cur)))
	     ((null? cur) #f)
	   (if (string=? (xmltrans:tag-ref #:name (car cur)) this-tag)
	       (exit (if (null? (cdr cur))
			 #f
			 (xmltrans:tag-ref #:name (cadr cur)))))))))
     ((pair? this-tag)
      (call-with-current-continuation
       (lambda (exit)
	 (do ((cur tag-list (cdr cur)))
	     ((null? cur) #f)
	   (if (eq? (car cur) this-tag)
	     (exit (if (null? (cdr cur))
		       #f
		       (xmltrans:tag-ref #:name (cadr cur)))))))))
     (else
      (xmltrans:tag-ref #:name (cadr tag-list))))))
    

	 

;;;; XML handlers

(define expected-next-tag #f)

(define (xmltrans:expect-next-tag tag-name thunk . rest)
  (set! expected-next-tag (list tag-name thunk rest)))

(define (start-element-handler name attrs) 
  (trace indent "ELEMENT START '" name "' - attributes " attrs)
  (indent-increase)
  (if (and expected-next-tag
	   (not (member name (list-ref expected-next-tag 2))))
      (begin
;	(xmltrans:parse-warning #f "expected tag:" expected-next-tag)
	(if (not (string=? (list-ref expected-next-tag 0) name))
	    (cond
	     ((list-ref  expected-next-tag 1) =>
	      (lambda (thunk)
		(thunk)))
	     (else
	      (xmltrans:parse-error #f "Expected " (car expected-next-tag))
	      (xmltrans:parse-error (cdr expected-next-tag)
				    "Previous element located here"))))
	(set! expected-next-tag #f)))
  (push-tag name attrs))

(define (end-element-handler name) 
  (indent-decrease)
  (trace indent "ELEMENT END '" name "'")
  (let ((text (pop-tag)))
    (cond
     ((null? tag-list)) ;; FIXME: Ignore text
     (else
      (push-text text)))))  

(define (data-handler value)
  (trace indent "CHARACTER DATA '" value "'")
  (push-text value))

(define (comment-handler value)
  (trace indent "COMMENT '" value "'"))

(define (start-cdata-section-handler) 
  (trace indent "CDATA START")
  (indent-increase))

(define (end-cdata-section-handler)
  (indent-decrease)
  (trace indent "CDATA END"))



(define input-file-name "<unknown file>")
(define input-line-number (lambda () 0))

(define-public (xmltrans:get-input-location)
  (cons input-file-name (input-line-number)))

(define-public (xmltrans:input-file-name)
  input-file-name)

(define-public (xmltrans:input-line-number)
  (input-line-number))

(define (xmltrans:parse-error locus . rest)
  (with-output-to-port
      (current-error-port)
    (lambda ()
      (let ((loc (or locus (xmltrans:get-input-location))))
	(display (car loc))
	(display ":")
	(display (cdr loc))
	(display ": ")
	(for-each
	 display
	 rest)
	(newline)))))

(define (xmltrans:parse-warning locus . rest)
  (with-output-to-port
      (current-error-port)
    (lambda ()
      (let ((loc (or locus (xmltrans:get-input-location))))
	(display (car loc))
	(display ":")
	(display (cdr loc))
	(display ": warning: ")
	(for-each
	 display
	 rest)
	(newline)))))

(define-public (xmltrans:dump-stack)
  (for-each
   (lambda (x)
     (display (xmltrans:tag-ref #:name x))
     (display " ")
     (display (xmltrans:tag-ref #:attrs x))
     (display " ")
     (display (xmltrans:tag-ref #:text x))
     (newline))
   tag-list))


;;;; Encoding
(define parser-encoding "UTF-8")

(define-public (xmltrans:set-input-encoding enc)
  (set! parser-encoding enc))

(define-public (xmltrans:input-encoding)
  parser-encoding)

 
;;;; Main procedure
(define-public (xmltrans:parse)
  (let ((parser (xml-make-parser
		 parser-encoding 
		 #:start-element-handler start-element-handler
		 #:end-element-handler end-element-handler
		 #:character-data-handler data-handler
		 #:comment-handler comment-handler
		 #:start-cdata-section-handler start-cdata-section-handler
		 #:end-cdata-section-handler end-cdata-section-handler)))

    (letrec ((get-line-pos (lambda ()
			     (xml-current-line-number parser))))
      (set! input-line-number get-line-pos)
      
      (catch 'gamma-xml-error
	     (lambda ()
	       (xml-parse parser))

	     (lambda (key subr fmt args ctx)
	       (let ((locus (cons input-file-name
				  (xml-error-descr ctx #:line))))
		 (cond
		  ((not ctx)
		   (xmltrans:parse-error locus
					 (apply format #t fmt args)))
		  (else
		   ; FIXME: (xml-error-descr ctx #:column)
		   (xmltrans:parse-error locus 
			 (xml-error-string (xml-error-descr ctx #:error-code)))
		   (if (xml-error-descr ctx #:has-context?)
		       (let ((ctx-text (xml-error-descr ctx #:context))
			     (ctx-pos  (xml-error-descr ctx #:error-offset)))
			 (format (current-error-port)
				 "Context (^ marks the point): ~A^~A~%"
				 (substring ctx-text 0 ctx-pos)
				 (substring ctx-text ctx-pos)))))))
	       (exit 1))))))

(define-public (xmltrans:parse-file file-name)
  (with-input-from-file
      file-name
      (lambda ()
	(let ((save-input-file-name input-file-name)
	      (save-input-line-number input-line-number))
	  (set! input-file-name file-name)
	  (xmltrans:parse)
	  (set! input-file-name save-input-file-name)
	  (set! input-line-number save-input-line-number)))))

;;;; xmltrans ends here



